/**************************************************
 Mex C Code for computing the local maximization
***************************************************/
# include <stdio.h>
# include <stdlib.h>
# include "mex.h"        /* the algorithm is connect to matlab */
# include "math.h"
# include "matrix.h"
# define PI 3.1415926
# define ABS(x) ((x)>0? (x):(-(x)))
# define MAX(x, y) ((x)>(y)? (x):(y))
# define MIN(x, y) ((x)<(y)? (x):(y))

/* getting the index of matlab image, (x,y) location, (sx, sy) sizes */
int px(int x, int y, int bx, int by)    
{            
   return (x + (y-1)*bx - 1); 
}

int *int_vector(int n)
{
    int *v; 
    v = (int*) mxCalloc (n, sizeof(int));
    return v;
}

int **int_matrix(int m, int n)
{
    int **mat; 
    int i; 
    
    mat = (int**) mxCalloc(m, sizeof(int*)); 
    for (i=0; i<m; i++)
        mat[i] = int_vector(n); 
    
    return mat; 
}

float *float_vector(int n)
{
    float *v; 
    v = (float*) mxCalloc (n, sizeof(float));
    return v;
}

void free_matrix(void **a, int nrow)
{
	int i;
	for(i = 0; i < nrow; i++)
		mxFree(a[i]);
	mxFree(a);
}

 /* variables */
int n;                      /* number of images */
int N;                      /* number of orientations */
float **fI;                 /* filtered images */
/* mxArray** fI; */
float **mI;                 /* locally maximized images */
int **sinsh, **cossh;       /* store the shift to avoid repeated sin and cosin computation */
int L, ore, sub;            /* allowed ranges of shifting in location and orientation */
int *sx, *sy;
 
/* store the shift values so that we do not need to repeat the sin and cos computation */ 
void storeshift()
{
    int ind, l;
    double theta;
    
    sinsh = int_matrix(N, L+L+1); 
    cossh = int_matrix(N, L+L+1);
    for (ind=0; ind<N; ind++)        
    {
        theta = PI*ind/N; 
        for (l=-L; l<=L; l++)
         {
            sinsh[ind][l+L] = floor(l*sub*sin(theta)+.5); 
            cossh[ind][l+L] = floor(l*sub*cos(theta)+.5); 
        }   
    }
}
    
/* for Gabor(x, y, orientation = ind), find local maximum in image i */
float shiftmax(int i, int ind, int x, int y, int *rm) 
{
   float m;
   int x1, y1, l, ml, mx, my, de, d, d1, here, mo, md;
 
   m = 0.;

   for (l=-L; l<=L; l++) 
   {
        x1 = x + cossh[ind][l+L]; 
        y1 = y + sinsh[ind][l+L];   /* shifting in normal direction */
        if ((x1>=1)&&(x1<=sx[i])&&(y1>=1)&&(y1<=sy[i]))
        {
        here = px(x1, y1, sx[i], sy[i]);
        for (de=-ore; de<=ore; de++)
            {
              d = de+ind;
              d1 = d;
              if (d<0)
                 d1 = d+N;
              if (d>=N)
                  d1 = d-N;   /* shifting in orientation */
              if (fI[d1*n+i][here]>m)
                {
                   m = fI[d1*n+i][here];   /* local maximization */
                   ml = l; mx = x1; my = y1; mo = d1; md = de; 
                }
             }
        }
     }
   rm[0] = ml; rm[1] = mx; rm[2] = my; rm[3] = mo; rm[4] = md; 
   return(m); 
}

 
 
void Cmm()
{
   int x, y, here, ind, i, ii, rm[5];
   int sx2, sy2;
   mI = mxCalloc(n*N,sizeof(*mI));
   for (i=0; i<n; i++)
   {
        for (ind=0; ind<N; ind++)
        {
           mI[ind*n+i] = float_vector( sx[i] * sy[i] ); /* will do subsampling in later versions */
           for (x=L*sub+1; x<sx[i]-L*sub-1; x++)
               for (y=L*sub+1; y<sy[i]-L*sub-1; y++)
               {
                   here = px(x, y, sx[i], sy[i]);
                   mI[ind*n+i][here] = shiftmax(i, ind, x, y, rm);
               }
        }
   } 
}
   


/* mex function is used to pass on the pointers and scalars from matlab, 
   so that heavy computation can be done by C, which puts the results into 
   some of the pointers. After that, matlab can then use these results. 
   
   So matlab is very much like a managing platform for organizing the 
   experiments, and mex C is like a work enginee for fast computation. */

void mexFunction(int nlhs, mxArray *plhs[], 
                 int nrhs, const mxArray *prhs[])                
{
    int ind, i;
    mwSize ndim;
    const mxArray *f;
    const mwSize* dims;
    mxClassID datatype;
    mwSize imSize[2];
    int count, x, y;
    float* pFloat;
    
    /*
     * Input from matlab
     * 0: allfiltered
     * 1: Lrange
     * 2: Orange
     * 3: sub
     */
    
    /* Get the number of dimensions in <allfiltered> */
    ndim = mxGetNumberOfDimensions(prhs[0]);
    dims = mxGetDimensions(prhs[0]);
    n = dims[0]; /* number of example images */
    N = dims[1]; /* number of orientations */
    sx = int_vector(n); sy = int_vector(n);
 
    fI = mxCalloc(n*N, sizeof(*fI));   /* fitered images */
    for (i=0; i<n; i++)
    {
        for (ind=0; ind<N; ind++)
        {
            f = mxGetCell(prhs[0], ind*n+i);
            datatype = mxGetClassID(f);
            if (datatype == mxDOUBLE_CLASS)
                printf("warning !!! not float \n");
            fI[ind*n+i] = (float*)(mxGetPr(f));    /* get pointers to filtered images */
        }
        sx[i] = mxGetM(f);
        sy[i] = mxGetN(f);
    }

    
    L = floor(mxGetScalar(prhs[1])+.5);     /* range of shifting along normal direction of Gabor */
    ore = floor(mxGetScalar(prhs[2])+.5);   /* range of shifting in orientation */
    sub = floor(mxGetScalar(prhs[3])+.5);   /* sub-sampling of Gabor filters to improve speed */

    storeshift();
    Cmm();

    plhs[0] = mxCreateCellArray(ndim, dims);
    for (i=0; i<n; i++)
    {
        for (ind=0; ind<N; ind++)
        {
            imSize[0] = sx[i];
            imSize[1] = sy[i];
            /* printf("imsize(1) = %d \n",imSize[0]); */
            f = mxCreateNumericArray(2,imSize,mxSINGLE_CLASS,0);
            pFloat = mxGetData(f);

            /*mxFree(f);*/
            count = 0;
            for( x = 0; x < sx[i]; x++)
                for (y=0; y < sy[i]; y++)
                {
                    pFloat[count] = mI[ind*n+i][count];
                    count++;
                }
            /* mxSetData(f,mI[ind*n+i]); */
            mxSetCell(plhs[0], ind*n+i, f);
        }
    }
}

     



 

                    